
(import scheme (chicken base) (chicken condition)
        pseudolists
        simple-tests)

(define-test (pseudolists?)
  (pl? "x")
  (pl? '(a b . c))
  (pl-null? 5)
  ((pl-of?) "x")
  ((pl-of? symbol?) '(a b . c))

  (equal? (pl-maker #f 1 2 3 4)
          '(1 2 3 4 . #f))
  (equal? (pl-maker #f 0 1 2 3) '(0 1 2 3 . #f))
  (equal? (pl-maker '() 0 1 2 3) '(0 1 2 3))
  (equal? (pl-iterate #f 5 add1 0) '(0 1 2 3 4 . #f))

  (not (pl-sentinel
         (pl-maker #f 1 2 3 4)))
  (= (pl-sentinel '(0 1 2 3 2 . 2)) 2)
  (not (pl-sentinel '(0 1 2 3 . #f)))
  (= (pl-length '(0 1 2 3 . 4)) 4)

  (equal? (pl-head '(0 1 2 3 2 . 2)) '(0 1 2 3 2))
  (equal? (pl-head 0) '())
  (equal? (pl-head '(0 . 1)) '(0))

  (= (pl-at 1 '(0 1 2 3 . 4)) 1)
  (= (pl-at 2 '(0 1 2 3 . #f)) 2)
  (not (condition-case (pl-at 0 1)
         ((exn) #f)))

  (= (pl-index odd? '(0 1 2 . 3)) 1)
  (= (pl-index odd? '(0 2 4 . 1)) -1)

  (equal? (pl-drop 1 '(0 1 2 3 . 4))
          '(1 2 3 . 4))
  (= (pl-drop 0 1) 1)
  (equal? (pl-drop 2 '(0 1 2 3 . #f)) '(2 3 . #f))
  (equal? (pl-drop-while odd? '(1 3 2 4 . #f)) '(2 4 . #f))
  (equal? (pl-drop-while negative? '(1 3 2 4 . #f)) '(1 3 2 4 . #f))
  
  (equal? (pl-take 3 '(0 1 2 3 4 . #t)) '(0 1 2 . #t))
  (equal? (pl-take 2 '(0 1 2 3 . #f)) '(0 1 . #f))
  (equal? (pl-take-while odd? '(1 3 2 4 . #f)) '(1 3 . #f))
  (not (pl-take-while negative? '(1 3 2 4 . #f)))

  (= (pl-filter odd? 1) 1)
  (equal? (pl-filter odd? '(0 1 2 3 4)) '(1 3))
  (equal? (pl-filter odd? '(0 1 2 3 . 4)) '(1 3 . 4))
  
  (equal? (pl-map add1 '(0 1 2 3 . 4)) '(1 2 3 4 . 4))
  (not (pl-map add1 #f))

  (equal? (pl-reverse '(0 1 2 3 . 4)) '(3 2 1 0 . 4))
  (equal? (pl-reverse '(0 1 2 3)) '(3 2 1 0))

  (equal? (pl-append '(0 1) '(2 3) #f) '(0 1 2 3 . #f))
  (equal? (pl-append '(0 1 . #f) '(2 3)) '(0 1 2 3))
  (equal? (pl-append '(0 1) '(2 3) '(4 5)) '(0 1 2 3 4 5))

  (= (pl-fold-right + 0 '(1 2 3 . #f)) 6)
  (= (pl-fold-left + 0 '(1 2 3)) 6)

  (equal? (pl-adjoin 2 '(0 1 2 3 . #f)) '(0 1 2 3 . #f))
  (equal? (pl-adjoin 4 '(0 1 2 3 #f)) '(4 0 1 2 3 #f))
  (equal? (pl-adjoin 1 '(0 1 2 3)) '(0 1 2 3))
  (equal? (pl-adjoin 1 '()) '(1))

  (equal? (pl-remove-dups '(0 1 2 3 2 . 2)) '(0 1 3 2 . 2))
  (equal? (pl-remove-dups '(0 1 2 3 2 2)) '(0 1 3 2))
  (equal? (pl-remove-dups '(0 1 2 1 3 2)) '(0 1 3 2))

  (equal? (pl-flatten '(0 1 . 2)) '(0 1 2))
  (equal? (pl-flatten '(0 (1 2))) '(0 1 2))
  (equal? (pl-flatten '(0 (1 (2 . 3)))) '(0 1 2 3))
  (equal? (pl-flatten '(0 (1 (2 . 3) 4))) '(0 1 2 3 4))
  (equal? (pl-flatten '(0 (1 (2 . 3) 4))) '(0 1 2 3 4))
  (equal? (pl-flatten '(0 (1 (2 . 3) 4 . #f))) '(0 1 2 3 4 #f))
  
  (equal? (pl-for ((x '(0 1 2 3 . #f))) (add1 x)) ; map
    '(1 2 3 4 . #f))
  (equal? (pl-for ((x '(0 1 2 3))) (add1 x)) ; map
    '(1 2 3 4))
  (equal? (pl-for ((x '(0 1 2 3 4 5 . #f) (odd? x))) x) ; filter
    '(1 3 5 . #f))
  (equal? (pl-for ((x '(0 1 2 3 4 5) (odd? x))) x) ; filter
    '(1 3 5))
  (equal? (pl-for ((n '(0 1 2 3 4 5) (positive? n) (even? n)))
            (* 10 n))
    '(20 40))
  (equal? (pl-for ((c '(A B C))
                   (k '(1 2 3 4)))
            (list c k))
    '((A 1) (A 2) (A 3) (A 4)
      (B 1) (B 2) (B 3) (B 4)
      (C 1) (C 2) (C 3) (C 4)))
  (equal? (pl-for ((c '(A B C . #f)) (k '(1 2 3 4 . #t))) (list c k))
    '((A 1) (A 2) (A 3) (A 4)
      (B 1) (B 2) (B 3) (B 4)
      (C 1) (C 2) (C 3) (C 4) . #f))
  )
 
(compound-test (PSEUDOLISTS)
  (pseudolists?)
  )
